#include <windows.h>
#include "kmp_pi.h"
#include "sap/sapLib.h"

#define KMPMODULE_PLUGIN_VERSION    2
#define KMPMODULE_PLUGIN_NAME       "m_sap.kpi v1.54r2 by Mamiya"
#define KMPMODULE_PLUGIN_COPYRIGHT  "SAP Library ver.1.54 by Adam Bienias"

#define KMPMODULE_REENTRANT 0
#define SIZE_OF_BUFFER 60
#define SAMPLE_RATE 44100	/* ύXs */
#define UNIT_RENDER (((SAMPLE_RATE*SIZE_OF_BUFFER)/1000) * 4)

class sapFile {
public:
	bool Open(const char *cszFileName, SOUNDINFO *pInfo);
	void SelectSong(int selectedSong);
	DWORD Write(BYTE *Buffer, DWORD dwSize);
	sapFile() : psap(0) {};
	~sapFile() { psap = 0; };
protected:
	sapMUSICstrc *psap;
	int curSong;
	int numOfSongs;
#define BYTES_OF_SKIPBUFFER ((SAMPLE_RATE/60) * 4)
	BYTE skipBuffer[BYTES_OF_SKIPBUFFER + 8];
};

bool sapFile::Open(const char *cszFileName, SOUNDINFO *pInfo)
{
	psap = sapLoadMusicFile((char *)cszFileName);
	if (!psap) return false;

	curSong = psap->defSong + 1;
	numOfSongs = psap->numOfSongs;
	if (numOfSongs < 1) numOfSongs = 1;

	pInfo->dwSamplesPerSec = SAMPLE_RATE;
	pInfo->dwChannels = 2;
	pInfo->dwBitsPerSample = 16;
	pInfo->dwSeekable = 1;
	pInfo->dwUnitRender = UNIT_RENDER;
	pInfo->dwReserved1 = 1;
	pInfo->dwReserved2 = 1;
	pInfo->dwLength = numOfSongs * 1000;

	SelectSong(1);

	return true;
}

void sapFile::SelectSong(int selectedSong)
{
	if (!psap) return;
	if (selectedSong) curSong = selectedSong;
	sapPlaySong(curSong - 1);
}

DWORD sapFile::Write(BYTE *Buffer, DWORD dwSize)
{
	DWORD pos = 0, len = dwSize;
	if (!dwSize || !psap) return 0;
	while (len > BYTES_OF_SKIPBUFFER)
	{
		sapRenderBuffer((short *)skipBuffer, BYTES_OF_SKIPBUFFER >> 2);
		if (Buffer) MoveMemory(&Buffer[pos], skipBuffer, BYTES_OF_SKIPBUFFER);
		pos += BYTES_OF_SKIPBUFFER;
		len -= BYTES_OF_SKIPBUFFER;
	}
	sapRenderBuffer((short *)skipBuffer, len >> 2);
	if (Buffer) MoveMemory(&Buffer[pos], skipBuffer, len);
	return dwSize;
}

#if !KMPMODULE_REENTRANT
static LONG mutex = 1;
#endif

static HKMP WINAPI kmp_Open(const char *cszFileName, SOUNDINFO *pInfo)
{
	LONG signal = 1;
	sapFile *pFile = 0;
#if !KMPMODULE_REENTRANT
	signal = InterlockedExchange(&mutex, 0);
#endif
	while (signal)
	{
		if (!pInfo) break;;
		pFile = new sapFile;
		if (!pFile) break;
		if (!pFile->Open(cszFileName, pInfo)) break;
		/* success */
		return (HKMP)pFile;
	}
	/* error */
	if (pFile) delete pFile;
#if !KMPMODULE_REENTRANT
	if (signal) InterlockedExchange(&mutex, signal);
#endif
	return 0;
}

static void WINAPI kmp_Close(HKMP hKMP)
{
	if (hKMP)
	{
		delete ((sapFile*)hKMP);
#if !KMPMODULE_REENTRANT
		InterlockedExchange(&mutex, 1);
#endif
	}
}

static DWORD WINAPI kmp_Render(HKMP hKMP, BYTE* Buffer, DWORD dwSize)
{
	if(!hKMP) return 0;
	DWORD dwWrite;
	dwWrite = ((sapFile*)hKMP)->Write(Buffer, dwSize);
#if 0
	/* Ȃ0ɂ(KvȂ?) */
	if (dwWrite && dwSize > dwWrite)
	{
		if (Buffer) memset(Buffer + dwWrite, 0, dwSize - dwWrite);
		dwWrite = dwSize;
	}
#endif
	return dwWrite;
}

static DWORD WINAPI kmp_SetPosition(HKMP hKMP, DWORD dwPos)
{
	if(!hKMP)return 0;
	((sapFile*)hKMP)->SelectSong(dwPos / 1000);
	return 0; /* tʒuԂ̂Ó? **/
	/* return dwPos; */
}

#if 0
static void WINAPI kmp_Init(void) {}
static void WINAPI kmp_Deinit(void) {}
static HKMP WINAPI kmp_OpenFromBuffer(const BYTE *Buffer, DWORD dwSize, SOUNDINFO *pInfo) { return 0; }
#endif

KMPMODULE* WINAPI kmp_GetTestModule(void)
{
	static const char *ppszSupportExts[] = { ".sap", NULL };
	static KMPMODULE Mod = {
		KMPMODULE_VERSION,
		KMPMODULE_PLUGIN_VERSION,
		KMPMODULE_PLUGIN_COPYRIGHT,
		KMPMODULE_PLUGIN_NAME,
		ppszSupportExts,
		KMPMODULE_REENTRANT,
		NULL,	//Init
		NULL,	//Deinit
		kmp_Open,
		NULL,	//OpenFromBuffer
		kmp_Close,
		kmp_Render,
		kmp_SetPosition
	};
	return &Mod;
}
